@@ -39,25 +39,39 @@ void sighandler(int foo)
 
 static int connect_cb(struct osmo_stream_cli *conn)
 {
-	LOGP(DSTREAMTEST, LOGL_NOTICE, "connected\n");
+	LOGP(DSTREAMTEST, LOGL_NOTICE, "connected: %s\n", osmo_stream_cli_get_sockname(conn));
+	return 0;
+}
+
+static int disconnect_cb(struct osmo_stream_cli *conn)
+{
+	LOGP(DSTREAMTEST, LOGL_NOTICE, "disconnected: %s\n", osmo_stream_cli_get_sockname(conn));
+
 	return 0;
 }
 
 static int read_cb(struct osmo_stream_cli *conn)
 {
+	int bytes;
 	struct msgb *msg;
 
-	LOGP(DSTREAMTEST, LOGL_NOTICE, "received message from stream\n");
+	LOGP(DSTREAMTEST, LOGL_NOTICE, "receiving message from stream... ");
 
 	msg = msgb_alloc(1024, "STREAMCLIENT/test");
 	if (msg == NULL) {
-		LOGP(DSTREAMTEST, LOGL_ERROR, "cannot allocate message\n");
+		LOGPC(DSTREAMTEST, LOGL_ERROR, "cannot allocate message\n");
 		return 0;
 	}
-	if (osmo_stream_cli_recv(conn, msg) < 0) {
-		LOGP(DSTREAMTEST, LOGL_ERROR, "cannot receive message\n");
+
+	bytes = osmo_stream_cli_recv(conn, msg);
+
+	if (bytes < 0) {
+		LOGPC(DSTREAMTEST, LOGL_ERROR, "cannot receive message\n");
 		return 0;
 	}
+
+	LOGPC(DSTREAMTEST, LOGL_NOTICE, "got %d (%d) bytes: %s\n", bytes, msg->len, msgb_hexdump(msg));
+
 	msgb_free(msg);
 	return 0;
 }
@@ -86,7 +100,7 @@ static int kbd_cb(struct osmo_fd *fd, unsigned int what)
 
 	osmo_stream_cli_send(conn, msg);
 
-	LOGP(DSTREAMTEST, LOGL_NOTICE, "message of %d bytes sent\n", msg->len);
+	LOGP(DSTREAMTEST, LOGL_NOTICE, "sent %d bytes message: %s\n", msg->len, msgb_hexdump(msg));
 
 	return 0;
 }
@@ -100,6 +114,7 @@ int main(void)
 	msgb_talloc_ctx_init(tall_test, 0);
 	osmo_init_logging2(tall_test, &osmo_stream_cli_test_log_info);
 	log_set_log_level(osmo_stderr_target, 1);
+	log_set_category_filter(osmo_stderr_target, DLINP, 0, LOGL_INFO);
 
 	/*
 	 * initialize stream cli.
@@ -113,6 +128,7 @@ int main(void)
 	osmo_stream_cli_set_addr(conn, "127.0.0.1");
 	osmo_stream_cli_set_port(conn, 10000);
 	osmo_stream_cli_set_connect_cb(conn, connect_cb);
+	osmo_stream_cli_set_disconnect_cb(conn, disconnect_cb);
 	osmo_stream_cli_set_read_cb(conn, read_cb);
 
 	if (osmo_stream_cli_open(conn) < 0) {
@@ -126,7 +142,7 @@ int main(void)
 		exit(EXIT_FAILURE);
 	}
 	kbd_ofd->fd = STDIN_FILENO;
-	kbd_ofd->when = BSC_FD_READ;
+	kbd_ofd->when = OSMO_FD_READ;
 	kbd_ofd->data = conn;
 	kbd_ofd->cb = kbd_cb;
 	rc = osmo_fd_register(kbd_ofd);
