@@ -146,7 +146,7 @@ osmux_rebuild_rtp(struct osmux_out_handle *h, struct osmux_hdr *osmuxh,
 	rtph->csrc_count = 0;
 	rtph->extension = 0;
 	rtph->version = RTP_VERSION;
-	rtph->payload_type = 98;
+	rtph->payload_type = h->rtp_payload_type;
 	/* ... emulate timestamp and ssrc */
 	rtph->timestamp = htonl(h->rtp_timestamp);
 	rtph->sequence = htons(h->rtp_seq);
@@ -366,8 +366,7 @@ static int osmux_batch_enqueue(struct msgb *msg, struct osmux_circuit *circuit,
 		if (rtph == NULL)
 			return -1;
 
-		LOGP(DLMUX, LOGL_DEBUG, "Batch is full for RTP "
-					"ssrc=%u\n", rtph->ssrc);
+		LOGP(DLMUX, LOGL_DEBUG, "Batch is full for RTP sssrc=%u\n", rtph->ssrc);
 		return -1;
 	}
 
@@ -999,16 +998,23 @@ osmux_tx_sched(struct llist_head *list,
 	}
 }
 
-void osmux_xfrm_output_init(struct osmux_out_handle *h, uint32_t rtp_ssrc)
+void osmux_xfrm_output_init2(struct osmux_out_handle *h, uint32_t rtp_ssrc, uint8_t rtp_payload_type)
 {
 	memset(h, 0, sizeof(*h));
 	h->rtp_seq = (uint16_t)random();
 	h->rtp_timestamp = (uint32_t)random();
 	h->rtp_ssrc = rtp_ssrc;
+	h->rtp_payload_type = rtp_payload_type;
 	INIT_LLIST_HEAD(&h->list);
 	osmo_timer_setup(&h->timer, osmux_xfrm_output_trigger, h);
 }
 
+void osmux_xfrm_output_init(struct osmux_out_handle *h, uint32_t rtp_ssrc)
+{
+	/* backward compatibility with old users, where 98 was harcoded in osmux_rebuild_rtp()  */
+	osmux_xfrm_output_init2(h, rtp_ssrc, 98);
+}
+
 #define SNPRINTF_BUFFER_SIZE(ret, remain, offset)	\
 	if (ret < 0)					\
 		ret = 0;				\
@@ -1025,7 +1031,7 @@ static int osmux_snprintf_header(char *buf, size_t size, struct osmux_hdr *osmux
 	ret = snprintf(buf, remain, "OSMUX seq=%03u ccid=%03u "
 				 "ft=%01u ctr=%01u "
 				 "amr_f=%01u amr_q=%01u "
-				 "amr_ft=%02u amr_cmr=%02u ",
+				 "amr_ft=%02u amr_cmr=%02u",
 			osmuxh->seq, osmuxh->circuit_id,
 			osmuxh->ft, osmuxh->ctr,
 			osmuxh->amr_f, osmuxh->amr_q,
@@ -1086,7 +1092,10 @@ int osmux_snprintf(char *buf, size_t size, struct msgb *msg)
 			return -1;
 		}
 		osmuxh = (struct osmux_hdr *)((uint8_t *)msg->data + msg_off);
-
+		if (msg_off) {
+			ret = snprintf(buf + offset, remain, ", ");
+			SNPRINTF_BUFFER_SIZE(ret, remain, offset);
+		}
 		ret = osmux_snprintf_header(buf + offset, remain, osmuxh);
 		SNPRINTF_BUFFER_SIZE(ret, remain, offset);
 
@@ -1116,6 +1125,8 @@ int osmux_snprintf(char *buf, size_t size, struct msgb *msg)
 			}
 
 			if (osmuxh->ft == OSMUX_FT_VOICE_AMR) {
+				ret = snprintf(buf + offset, remain, " ");
+				SNPRINTF_BUFFER_SIZE(ret, remain, offset);
 				ret = osmux_snprintf_payload(buf + offset, remain,
 							     osmux_get_payload(osmuxh),
 							     payload_len);
