@@ -49,6 +49,9 @@
 #include <netinet/sctp.h>
 #endif
 
+#define LOGSCLI(cli, level, fmt, args...) \
+	LOGP(DLINP, level, "[%s] %s(): " fmt, get_value_string(stream_cli_state_names, (cli)->state), __func__, ## args)
+
 /*! \addtogroup stream Osmocom Stream Socket
  *  @{
  *
@@ -62,6 +65,7 @@
  *  \brief Osmocom stream socket helpers
  */
 
+#ifdef HAVE_LIBSCTP
 /*
  * Platforms that don't have MSG_NOSIGNAL (which disables SIGPIPE)
  * usually have SO_NOSIGPIPE (set via setsockopt).
@@ -70,6 +74,99 @@
 #define MSG_NOSIGNAL 0
 #endif
 
+/* is any of the bytes from offset .. u8_size in 'u8' non-zero? return offset or -1 if all zero */
+static int byte_nonzero(const uint8_t *u8, unsigned int offset, unsigned int u8_size)
+{
+	int j;
+
+	for (j = offset; j < u8_size; j++) {
+		if (u8[j] != 0)
+			return j;
+	}
+
+	return -1;
+}
+
+static int sctp_sockopt_event_subscribe_size = 0;
+
+static int determine_sctp_sockopt_event_subscribe_size(void)
+{
+	uint8_t buf[256];
+	socklen_t buf_len = sizeof(buf);
+	int sd, rc;
+
+	/* only do this once */
+	if (sctp_sockopt_event_subscribe_size != 0)
+		return 0;
+
+	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_SCTP);
+	if (sd < 0)
+		return sd;
+
+	rc = getsockopt(sd, IPPROTO_SCTP, SCTP_EVENTS, buf, &buf_len);
+	close(sd);
+	if (rc < 0)
+		return rc;
+
+	sctp_sockopt_event_subscribe_size = buf_len;
+
+	LOGP(DLINP, LOGL_INFO, "sizes of 'struct sctp_event_subscribe': compile-time %zu, kernel: %u\n",
+		sizeof(struct sctp_event_subscribe), sctp_sockopt_event_subscribe_size);
+	return 0;
+}
+
+/* Attempt to work around Linux kernel ABI breakage
+ *
+ * The Linux kernel ABI for the SCTP_EVENTS socket option has been broken repeatedly.
+ *  - until commit 35ea82d611da59f8bea44a37996b3b11bb1d3fd7 ( kernel < 4.11), the size is 10 bytes
+ *  - in 4.11 it is 11 bytes
+ *  - in 4.12 .. 5.4 it is 13 bytes
+ *  - in kernels >= 5.5 it is 14 bytes
+ *
+ * This wouldn't be a problem if the kernel didn't have a "stupid" assumption that the structure
+ * size passed by userspace will match 1:1 the length of the structure at kernel compile time. In
+ * an ideal world, it would just use the known first bytes and assume the remainder is all zero.
+ * But as it doesn't do that, let's try to work around this */
+static int sctp_setsockopt_events_linux_workaround(int fd, const struct sctp_event_subscribe *event)
+{
+
+	const unsigned int compiletime_size = sizeof(*event);
+	int rc;
+
+	if (determine_sctp_sockopt_event_subscribe_size() < 0) {
+		LOGP(DLINP, LOGL_ERROR, "Cannot determine SCTP_EVENTS socket option size\n");
+		return -1;
+	}
+
+	if (compiletime_size == sctp_sockopt_event_subscribe_size) {
+		/* no kernel workaround needed */
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, event, compiletime_size);
+	} else if (compiletime_size < sctp_sockopt_event_subscribe_size) {
+		/* we are using an older userspace with a more modern kernel and hence need
+		 * to pad the data */
+		uint8_t buf[sctp_sockopt_event_subscribe_size];
+
+		memcpy(buf, event, compiletime_size);
+		memset(buf + sizeof(*event), 0, sctp_sockopt_event_subscribe_size - compiletime_size);
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, buf, sctp_sockopt_event_subscribe_size);
+	} else /* if (compiletime_size > sctp_sockopt_event_subscribe_size) */ {
+		/* we are using a newer userspace with an older kernel and hence need to truncate
+		 * the data - but only if the caller didn't try to enable any of the events of the
+		 * truncated portion */
+		rc = byte_nonzero((const uint8_t *)event, sctp_sockopt_event_subscribe_size,
+				  compiletime_size);
+		if (rc >= 0) {
+			LOGP(DLINP, LOGL_ERROR, "Kernel only supports sctp_event_subscribe of %u bytes, "
+				"but caller tried to enable more modern event at offset %u\n",
+				sctp_sockopt_event_subscribe_size, rc);
+			return -1;
+		}
+
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, event, sctp_sockopt_event_subscribe_size);
+	}
+}
+#endif // HAVE_LIBSCTP
+
 static int sctp_sock_activate_events(int fd)
 {
 #ifdef HAVE_LIBSCTP
@@ -81,18 +178,15 @@ static int sctp_sock_activate_events(int fd)
 	event.sctp_data_io_event = 1;
 	event.sctp_association_event = 1;
 	event.sctp_address_event = 1;
-	event.sctp_address_event = 1;
 	event.sctp_send_failure_event = 1;
 	event.sctp_peer_error_event = 1;
 	event.sctp_shutdown_event = 1;
 	/* IMPORTANT: Do NOT enable sender_dry_event here, see
 	 * https://bugzilla.redhat.com/show_bug.cgi?id=1442784 */
-	rc = setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS,
-			&event, sizeof(event));
 
+	rc = sctp_setsockopt_events_linux_workaround(fd, &event);
 	if (rc < 0)
-		LOGP(DLINP, LOGL_ERROR, "coudldn't activate SCTP events "
-		     "on FD %u\n", fd);
+		LOGP(DLINP, LOGL_ERROR, "couldn't activate SCTP events on FD %u\n", fd);
 	return rc;
 #else
 	return -1;
@@ -104,9 +198,11 @@ static int setsockopt_nodelay(int fd, int proto, int on)
 	int rc;
 
 	switch (proto) {
+#ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
 		rc = setsockopt(fd, IPPROTO_SCTP, SCTP_NODELAY, &on, sizeof(on));
 		break;
+#endif
 	case IPPROTO_TCP:
 		rc = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
 		break;
@@ -125,26 +221,44 @@ static int setsockopt_nodelay(int fd, int proto, int on)
  */
 
 enum osmo_stream_cli_state {
-        STREAM_CLI_STATE_NONE         = 0,
-        STREAM_CLI_STATE_CONNECTING   = 1,
-        STREAM_CLI_STATE_CONNECTED    = 2,
+        STREAM_CLI_STATE_CLOSED,	 /* No fd associated, no timer active */
+        STREAM_CLI_STATE_WAIT_RECONNECT, /* No fd associated, has timer active to try to connect again */
+        STREAM_CLI_STATE_CONNECTING,	 /* Fd associated, but connection not yet confirmed by peer or lower layers */
+        STREAM_CLI_STATE_CONNECTED,	 /* Fd associated and connection is established */
         STREAM_CLI_STATE_MAX
 };
 
+static const struct value_string stream_cli_state_names[] = {
+	{ STREAM_CLI_STATE_CLOSED,         "CLOSED" },
+	{ STREAM_CLI_STATE_WAIT_RECONNECT, "WAIT_RECONNECT" },
+	{ STREAM_CLI_STATE_CONNECTING,     "CONNECTING" },
+	{ STREAM_CLI_STATE_CONNECTED,      "CONNECTED" },
+	{ 0, NULL }
+};
+
 #define OSMO_STREAM_CLI_F_RECONF	(1 << 0)
 #define OSMO_STREAM_CLI_F_NODELAY	(1 << 1)
 
+#ifdef HAVE_LIBSCTP
+#define OSMO_STREAM_MAX_ADDRS OSMO_SOCK_MAX_ADDRS
+#else
+#define OSMO_STREAM_MAX_ADDRS 1
+#endif
+
 struct osmo_stream_cli {
 	struct osmo_fd			ofd;
 	struct llist_head		tx_queue;
 	struct osmo_timer_list		timer;
 	enum osmo_stream_cli_state	state;
-	char				*addr;
+	char				*addr[OSMO_STREAM_MAX_ADDRS];
+	uint8_t                         addrcnt;
 	uint16_t			port;
-	char				*local_addr;
+	char				*local_addr[OSMO_STREAM_MAX_ADDRS];
+	uint8_t                         local_addrcnt;
 	uint16_t			local_port;
 	uint16_t			proto;
 	int (*connect_cb)(struct osmo_stream_cli *srv);
+	int (*disconnect_cb)(struct osmo_stream_cli *srv);
 	int (*read_cb)(struct osmo_stream_cli *srv);
 	int (*write_cb)(struct osmo_stream_cli *srv);
 	void				*data;
@@ -155,20 +269,30 @@ struct osmo_stream_cli {
 void osmo_stream_cli_close(struct osmo_stream_cli *cli);
 
 /*! \brief Re-connect an Osmocom Stream Client
- *  If re-connection is enabled for this client, we close any existing
- *  connection (if any) and schedule a re-connect timer */
+ *  If re-connection is enabled for this client
+ *  (which is the case unless negative timeout was explicitly set via osmo_stream_cli_set_reconnect_timeout() call),
+ *  we close any existing connection (if any) and schedule a re-connect timer */
 void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
 {
+	osmo_stream_cli_close(cli);
+
 	if (cli->reconnect_timeout < 0) {
-		LOGP(DLINP, LOGL_DEBUG, "not reconnecting, disabled.\n");
+		LOGSCLI(cli, LOGL_INFO, "not reconnecting, disabled.\n");
 		return;
 	}
-	LOGP(DLINP, LOGL_DEBUG, "connection closed\n");
-	osmo_stream_cli_close(cli);
-	LOGP(DLINP, LOGL_DEBUG, "retrying in %d seconds...\n",
+
+	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
+	LOGSCLI(cli, LOGL_INFO, "retrying in %d seconds...\n",
 		cli->reconnect_timeout);
 	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
-	cli->state = STREAM_CLI_STATE_CONNECTING;
+}
+
+/*! \brief Check if Osmocom Stream Client is in connected state
+ *  \param[in] cli Osmocom Stream Client
+ */
+bool osmo_stream_cli_is_connected(struct osmo_stream_cli *cli)
+{
+	return cli->state == STREAM_CLI_STATE_CONNECTED;
 }
 
 /*! \brief Close an Osmocom Stream Client
@@ -182,11 +306,19 @@ void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 	osmo_fd_unregister(&cli->ofd);
 	close(cli->ofd.fd);
 	cli->ofd.fd = -1;
+
+	if (cli->state == STREAM_CLI_STATE_CONNECTED) {
+		LOGSCLI(cli, LOGL_DEBUG, "connection closed\n");
+		if (cli->disconnect_cb)
+			cli->disconnect_cb(cli);
+	}
+
+	cli->state = STREAM_CLI_STATE_CLOSED;
 }
 
 static void osmo_stream_cli_read(struct osmo_stream_cli *cli)
 {
-	LOGP(DLINP, LOGL_DEBUG, "message received\n");
+	LOGSCLI(cli, LOGL_DEBUG, "message received\n");
 
 	if (cli->read_cb)
 		cli->read_cb(cli);
@@ -201,21 +333,21 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 	struct llist_head *lh;
 	int ret;
 
-	LOGP(DLINP, LOGL_DEBUG, "sending data\n");
-
 	if (llist_empty(&cli->tx_queue)) {
-		cli->ofd.when &= ~BSC_FD_WRITE;
+		cli->ofd.when &= ~OSMO_FD_WRITE;
 		return 0;
 	}
 	lh = cli->tx_queue.next;
 	llist_del(lh);
 	msg = llist_entry(lh, struct msgb, list);
 
-	if (cli->state == STREAM_CLI_STATE_CONNECTING) {
-		LOGP(DLINP, LOGL_ERROR, "not connected, dropping data!\n");
+	if (!osmo_stream_cli_is_connected(cli)) {
+		LOGSCLI(cli, LOGL_ERROR, "not connected, dropping data!\n");
 		return 0;
 	}
 
+	LOGSCLI(cli, LOGL_DEBUG, "sending %u bytes of data\n", msgb_length(msg));
+
 	switch (cli->proto) {
 #ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
@@ -228,14 +360,14 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 #endif
 	case IPPROTO_TCP:
 	default:
-		ret = send(cli->ofd.fd, msg->data, msg->len, 0);
+		ret = send(cli->ofd.fd, msg->data, msgb_length(msg), 0);
 		break;
 	}
 	if (ret < 0) {
 		if (errno == EPIPE || errno == ENOTCONN) {
 			osmo_stream_cli_reconnect(cli);
 		}
-		LOGP(DLINP, LOGL_ERROR, "error to send\n");
+		LOGSCLI(cli, LOGL_ERROR, "error %d to send\n", ret);
 	}
 	msgb_free(msg);
 	return 0;
@@ -254,8 +386,13 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 			osmo_stream_cli_reconnect(cli);
 			return 0;
 		}
-		ofd->when &= ~BSC_FD_WRITE;
-		LOGP(DLINP, LOGL_DEBUG, "connection done.\n");
+
+		/* If messages got enqueued while 'connecting', keep WRITE flag
+		   up to dispatch them upon next main loop step */
+		if (llist_empty(&cli->tx_queue))
+			cli->ofd.when &= ~OSMO_FD_WRITE;
+
+		LOGSCLI(cli, LOGL_DEBUG, "connection done.\n");
 		cli->state = STREAM_CLI_STATE_CONNECTED;
 		if (cli->proto == IPPROTO_SCTP) {
 #ifdef SO_NOSIGPIPE
@@ -263,7 +400,7 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 
 			ret = setsockopt(ofd->fd, SOL_SOCKET, SO_NOSIGPIPE, (void*)&val, sizeof(val));
 			if (ret < 0)
-				LOGP(DLINP, LOGL_DEBUG, "Failed setting SO_NOSIGPIPE: %s\n", strerror(errno));
+				LOGSCLI(cli, LOGL_DEBUG, "Failed setting SO_NOSIGPIPE: %s\n", strerror(errno));
 #endif
 			sctp_sock_activate_events(ofd->fd);
 		}
@@ -271,17 +408,18 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 			cli->connect_cb(cli);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
-		if (what & BSC_FD_READ) {
-			LOGP(DLINP, LOGL_DEBUG, "connected read\n");
+		if (what & OSMO_FD_READ) {
+			LOGSCLI(cli, LOGL_DEBUG, "connected read\n");
 			osmo_stream_cli_read(cli);
 		}
-		if (what & BSC_FD_WRITE) {
-			LOGP(DLINP, LOGL_DEBUG, "connected write\n");
+		if (what & OSMO_FD_WRITE) {
+			LOGSCLI(cli, LOGL_DEBUG, "connected write\n");
 			osmo_stream_cli_write(cli);
 		}
 		break;
 	default:
-		break;
+		/* Only CONNECTING and CONNECTED states are expected, since they are the only states where FD exists: */
+		osmo_panic("osmo_stream_cli_fd_cb called with unexpected state %d\n", cli->state);
 	}
         return 0;
 }
@@ -302,11 +440,10 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 
 	cli->proto = IPPROTO_TCP;
 	cli->ofd.fd = -1;
-	cli->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
 	cli->ofd.priv_nr = 0;	/* XXX */
 	cli->ofd.cb = osmo_stream_cli_fd_cb;
 	cli->ofd.data = cli;
-	cli->state = STREAM_CLI_STATE_CONNECTING;
+	cli->state = STREAM_CLI_STATE_CLOSED;
 	osmo_timer_setup(&cli->timer, cli_timer_cb, cli);
 	cli->reconnect_timeout = 5;	/* default is 5 seconds. */
 	INIT_LLIST_HEAD(&cli->tx_queue);
@@ -321,8 +458,32 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 void
 osmo_stream_cli_set_addr(struct osmo_stream_cli *cli, const char *addr)
 {
-	osmo_talloc_replace_string(cli, &cli->addr, addr);
+	osmo_stream_cli_set_addrs(cli, &addr, 1);
+}
+
+/*! \brief Set the remote address set to which we connect.
+ *  Useful for protocols allowing connecting to more than one address (such as SCTP)
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] addr Remote IP address set
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_addrs(struct osmo_stream_cli *cli, const char **addr, size_t addrcnt)
+{
+	int i = 0;
+
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
+		return -EINVAL;
+
+	for (; i < addrcnt; i++)
+		osmo_talloc_replace_string(cli, &cli->addr[i], addr[i]);
+	for (; i < cli->addrcnt; i++) {
+			talloc_free(cli->addr[i]);
+			cli->addr[i] = NULL;
+	}
+
+	cli->addrcnt = addrcnt;
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
+	return 0;
 }
 
 /*! \brief Set the remote port number to which we connect
@@ -354,8 +515,32 @@ osmo_stream_cli_set_local_port(struct osmo_stream_cli *cli, uint16_t port)
 void
 osmo_stream_cli_set_local_addr(struct osmo_stream_cli *cli, const char *addr)
 {
-	osmo_talloc_replace_string(cli, &cli->local_addr, addr);
+	osmo_stream_cli_set_local_addrs(cli, &addr, 1);
+}
+
+/*! \brief Set the local address set to which we connect.
+ *  Useful for protocols allowing bind to more than one address (such as SCTP)
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] addr Local IP address set
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_local_addrs(struct osmo_stream_cli *cli, const char **addr, size_t addrcnt)
+{
+	int i = 0;
+
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
+		return -EINVAL;
+
+	for (; i < addrcnt; i++)
+		osmo_talloc_replace_string(cli, &cli->local_addr[i], addr[i]);
+	for (; i < cli->local_addrcnt; i++) {
+			talloc_free(cli->local_addr[i]);
+			cli->local_addr[i] = NULL;
+	}
+
+	cli->local_addrcnt = addrcnt;
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
+	return 0;
 }
 
 /*! \brief Set the protocol for the stream client socket
@@ -371,7 +556,7 @@ osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto)
 
 /*! \brief Set the reconnect time of the stream client socket
  *  \param[in] cli Stream Client to modify
- *  \param[in] timeout Re-connect timeout in seconds */
+ *  \param[in] timeout Re-connect timeout in seconds or negative value to disable auto-reconnection */
 void
 osmo_stream_cli_set_reconnect_timeout(struct osmo_stream_cli *cli, int timeout)
 {
@@ -388,15 +573,27 @@ osmo_stream_cli_set_data(struct osmo_stream_cli *cli, void *data)
 }
 
 /*! \brief Get application private data of the stream client socket
- *  \param[in] cli Stream Client to modif
+ *  \param[in] cli Stream Client to modify
  *  \returns Application private data, as set by \ref osmo_stream_cli_set_data() */
 void *osmo_stream_cli_get_data(struct osmo_stream_cli *cli)
 {
 	return cli->data;
 }
 
+/*! \brief Get the stream client socket description.
+ *  \param[in] cli Stream Client to examine
+ *  \returns Socket description or NULL in case of error */
+char *osmo_stream_cli_get_sockname(const struct osmo_stream_cli *cli)
+{
+	static char buf[OSMO_SOCK_NAME_MAXLEN];
+
+	osmo_sock_get_name_buf(buf, OSMO_SOCK_NAME_MAXLEN, cli->ofd.fd);
+
+	return buf;
+}
+
 /*! \brief Get Osmocom File Descriptor of the stream client socket
- *  \param[in] cli Stream Client to modif
+ *  \param[in] cli Stream Client to modify
  *  \returns Pointer to \ref osmo_fd */
 struct osmo_fd *
 osmo_stream_cli_get_ofd(struct osmo_stream_cli *cli)
@@ -405,7 +602,7 @@ osmo_stream_cli_get_ofd(struct osmo_stream_cli *cli)
 }
 
 /*! \brief Set the call-back function called on connect of the stream client socket
- *  \param[in] cli Stream Client to modif
+ *  \param[in] cli Stream Client to modify
  *  \param[in] connect_cb Call-back function to be called upon connect */
 void
 osmo_stream_cli_set_connect_cb(struct osmo_stream_cli *cli,
@@ -414,8 +611,17 @@ osmo_stream_cli_set_connect_cb(struct osmo_stream_cli *cli,
 	cli->connect_cb = connect_cb;
 }
 
+/*! \brief Set the call-back function called on disconnect of the stream client socket
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] disconnect_cb Call-back function to be called upon disconnect */
+void osmo_stream_cli_set_disconnect_cb(struct osmo_stream_cli *cli,
+				       int (*disconnect_cb)(struct osmo_stream_cli *cli))
+{
+	cli->disconnect_cb = disconnect_cb;
+}
+
 /*! \brief Set the call-back function called to read from the stream client socket
- *  \param[in] cli Stream Client to modif
+ *  \param[in] cli Stream Client to modify
  *  \param[in] read_cb Call-back function to be called when we want to read */
 void
 osmo_stream_cli_set_read_cb(struct osmo_stream_cli *cli,
@@ -434,7 +640,8 @@ void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
 	talloc_free(cli);
 }
 
-/*! \brief Open connection of an Osmocom stream client
+/*! \brief DEPRECATED: use osmo_stream_cli_set_reconnect_timeout() or osmo_stream_cli_reconnect() instead!
+ * Open connection of an Osmocom stream client
  *  \param[in] cli Stream Client to connect
  *  \param[in] reconect 1 if we should not automatically reconnect
  */
@@ -448,16 +655,29 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
 
-	ret = osmo_sock_init2(AF_INET, SOCK_STREAM, cli->proto,
-			      cli->local_addr, cli->local_port,
-			      cli->addr, cli->port,
-			      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+	switch (cli->proto) {
+#ifdef HAVE_LIBSCTP
+	case IPPROTO_SCTP:
+		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, cli->proto,
+						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
+						(const char **)cli->addr, cli->addrcnt, cli->port,
+						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+		break;
+#endif
+	default:
+		ret = osmo_sock_init2(AF_INET, SOCK_STREAM, cli->proto,
+				      cli->local_addr[0], cli->local_port,
+				      cli->addr[0], cli->port,
+				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+	}
+
 	if (ret < 0) {
 		if (reconnect)
 			osmo_stream_cli_reconnect(cli);
 		return ret;
 	}
 	cli->ofd.fd = ret;
+	cli->ofd.when = OSMO_FD_READ | OSMO_FD_WRITE;
 
 	if (cli->flags & OSMO_STREAM_CLI_F_NODELAY) {
 		ret = setsockopt_nodelay(cli->ofd.fd, cli->proto, 1);
@@ -468,10 +688,11 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 	if (osmo_fd_register(&cli->ofd) < 0)
 		goto error_close_socket;
 
+	cli->state = STREAM_CLI_STATE_CONNECTING;
 	return 0;
 
 error_close_socket:
-	close(ret);
+	close(cli->ofd.fd);
 	cli->ofd.fd = -1;
 	return -EIO;
 }
@@ -493,26 +714,68 @@ void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay)
 }
 
 /*! \brief Open connection of an Osmocom stream client
+ *  By default the client will automatically reconnect after default timeout.
+ *  To disable this, use osmo_stream_cli_set_reconnect_timeout() before calling this function.
  *  \param[in] cli Stream Client to connect */
 int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 {
-	return osmo_stream_cli_open2(cli, 0);
+	int ret;
+
+	/* we are reconfiguring this socket, close existing first. */
+	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && cli->ofd.fd >= 0)
+		osmo_stream_cli_close(cli);
+
+	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
+
+
+	switch (cli->proto) {
+#ifdef HAVE_LIBSCTP
+	case IPPROTO_SCTP:
+		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, cli->proto,
+						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
+						(const char **)cli->addr, cli->addrcnt, cli->port,
+						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+		break;
+#endif
+	default:
+		ret = osmo_sock_init2(AF_INET, SOCK_STREAM, cli->proto,
+				      cli->local_addr[0], cli->local_port,
+				      cli->addr[0], cli->port,
+				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+	}
+
+	if (ret < 0) {
+		osmo_stream_cli_reconnect(cli);
+		return ret;
+	}
+	cli->ofd.fd = ret;
+	cli->ofd.when = OSMO_FD_READ | OSMO_FD_WRITE;
+
+	if (cli->flags & OSMO_STREAM_CLI_F_NODELAY) {
+		ret = setsockopt_nodelay(cli->ofd.fd, cli->proto, 1);
+		if (ret < 0)
+			goto error_close_socket;
+	}
+
+	if (osmo_fd_register(&cli->ofd) < 0)
+		goto error_close_socket;
+
+	cli->state = STREAM_CLI_STATE_CONNECTING;
+	return 0;
+
+error_close_socket:
+	cli->state = STREAM_CLI_STATE_CLOSED;
+	close(cli->ofd.fd);
+	cli->ofd.fd = -1;
+	return -EIO;
 }
 
 static void cli_timer_cb(void *data)
 {
 	struct osmo_stream_cli *cli = data;
 
-	LOGP(DLINP, LOGL_DEBUG, "reconnecting.\n");
-
-	switch(cli->state) {
-	case STREAM_CLI_STATE_CONNECTING:
-		cli->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
-		osmo_stream_cli_open2(cli, 1);
-	        break;
-	default:
-		break;
-	}
+	LOGSCLI(cli, LOGL_DEBUG, "reconnecting.\n");
+	osmo_stream_cli_open(cli);
 }
 
 /*! \brief Enqueue data to be sent via an Osmocom stream client
@@ -521,7 +784,7 @@ static void cli_timer_cb(void *data)
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 {
 	msgb_enqueue(&cli->tx_queue, msg);
-	cli->ofd.when |= BSC_FD_WRITE;
+	cli->ofd.when |= OSMO_FD_WRITE;
 }
 
 /*! \brief Receive data via an Osmocom stream client
@@ -535,18 +798,17 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	ret = recv(cli->ofd.fd, msg->data, msg->data_len, 0);
 	if (ret < 0) {
 		if (errno == EPIPE || errno == ECONNRESET) {
-			LOGP(DLINP, LOGL_ERROR,
-				"lost connection with srv\n");
+			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv\n");
 		}
 		osmo_stream_cli_reconnect(cli);
 		return ret;
 	} else if (ret == 0) {
-		LOGP(DLINP, LOGL_ERROR, "connection closed with srv\n");
+		LOGSCLI(cli, LOGL_ERROR, "connection closed with srv\n");
 		osmo_stream_cli_reconnect(cli);
 		return ret;
 	}
 	msgb_put(msg, ret);
-	LOGP(DLINP, LOGL_DEBUG, "received %d bytes from srv\n", ret);
+	LOGSCLI(cli, LOGL_DEBUG, "received %d bytes from srv\n", ret);
 	return ret;
 }
 
@@ -559,7 +821,8 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 
 struct osmo_stream_srv_link {
         struct osmo_fd                  ofd;
-        char                            *addr;
+        char                            *addr[OSMO_STREAM_MAX_ADDRS];
+        uint8_t                         addrcnt;
         uint16_t                        port;
         uint16_t                        proto;
         int (*accept_cb)(struct osmo_stream_srv_link *srv, int fd);
@@ -628,7 +891,7 @@ struct osmo_stream_srv_link *osmo_stream_srv_link_create(void *ctx)
 
 	link->proto = IPPROTO_TCP;
 	link->ofd.fd = -1;
-	link->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
+	link->ofd.when |= OSMO_FD_READ | OSMO_FD_WRITE;
 	link->ofd.cb = osmo_stream_srv_fd_cb;
 	link->ofd.data = link;
 
@@ -657,8 +920,32 @@ void osmo_stream_srv_link_set_nodelay(struct osmo_stream_srv_link *link, bool no
 void osmo_stream_srv_link_set_addr(struct osmo_stream_srv_link *link,
 				      const char *addr)
 {
-	osmo_talloc_replace_string(link, &link->addr, addr);
+	osmo_stream_srv_link_set_addrs(link, &addr, 1);
+}
+
+/*! \brief Set the local address set to which we bind.
+ *  Useful for protocols allowing bind on more than one address (such as SCTP)
+ *  \param[in] link Stream Server Link to modify
+ *  \param[in] addr Local IP address
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_srv_link_set_addrs(struct osmo_stream_srv_link *link, const char **addr, size_t addrcnt)
+{
+	int i = 0;
+
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
+		return -EINVAL;
+
+	for (; i < addrcnt; i++)
+		osmo_talloc_replace_string(link, &link->addr[i], addr[i]);
+	for (; i < link->addrcnt; i++) {
+			talloc_free(link->addr[i]);
+			link->addr[i] = NULL;
+	}
+
+	link->addrcnt = addrcnt;
 	link->flags |= OSMO_STREAM_SRV_F_RECONF;
+	return 0;
 }
 
 /*! \brief Set the local port number to which we bind
@@ -702,6 +989,25 @@ void *osmo_stream_srv_link_get_data(struct osmo_stream_srv_link *link)
 	return link->data;
 }
 
+/*! \brief Get description of the stream server link e. g. 127.0.0.1:1234
+ *  \param[in] link Stream Server Link to examine
+ *  \returns Link description or NULL in case of error */
+char *osmo_stream_srv_link_get_sockname(const struct osmo_stream_srv_link *link)
+{
+	static char buf[INET6_ADDRSTRLEN + 6];
+	int rc = osmo_sock_get_local_ip(link->ofd.fd, buf, INET6_ADDRSTRLEN);
+	if (rc < 0)
+		return NULL;
+
+	buf[strnlen(buf, INET6_ADDRSTRLEN + 6)] = ':';
+
+	rc = osmo_sock_get_local_ip_port(link->ofd.fd, buf + strnlen(buf, INET6_ADDRSTRLEN + 6), 6);
+	if (rc < 0)
+		return NULL;
+
+	return buf;
+}
+
 /*! \brief Get Osmocom File Descriptor of the stream server link
  *  \param[in] link Stream Server Link
  *  \returns Pointer to \ref osmo_fd */
@@ -746,8 +1052,18 @@ int osmo_stream_srv_link_open(struct osmo_stream_srv_link *link)
 
 	link->flags &= ~OSMO_STREAM_SRV_F_RECONF;
 
-	ret = osmo_sock_init(AF_INET, SOCK_STREAM, link->proto,
-			     link->addr, link->port, OSMO_SOCK_F_BIND);
+	switch (link->proto) {
+#ifdef HAVE_LIBSCTP
+	case IPPROTO_SCTP:
+		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, link->proto,
+						(const char **)link->addr, link->addrcnt, link->port,
+						NULL, 0, 0, OSMO_SOCK_F_BIND);
+		break;
+#endif
+	default:
+		ret = osmo_sock_init(AF_INET, SOCK_STREAM, link->proto,
+					link->addr[0], link->port, OSMO_SOCK_F_BIND);
+	}
 	if (ret < 0)
 		return ret;
 
@@ -784,19 +1100,21 @@ struct osmo_stream_srv {
 	int				flags;
 };
 
-static void osmo_stream_srv_read(struct osmo_stream_srv *conn)
+static int osmo_stream_srv_read(struct osmo_stream_srv *conn)
 {
+	int rc = 0;
+
 	LOGP(DLINP, LOGL_DEBUG, "message received\n");
 
 	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
 		LOGP(DLINP, LOGL_DEBUG, "Connection is being flushed and closed; ignoring received message\n");
-		return;
+		return 0;
 	}
 
 	if (conn->cb)
-		conn->cb(conn);
+		rc = conn->cb(conn);
 
-	return;
+	return rc;
 }
 
 static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
@@ -811,7 +1129,7 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 	LOGP(DLINP, LOGL_DEBUG, "sending data\n");
 
 	if (llist_empty(&conn->tx_queue)) {
-		conn->ofd.when &= ~BSC_FD_WRITE;
+		conn->ofd.when &= ~OSMO_FD_WRITE;
 		return;
 	}
 	lh = conn->tx_queue.next;
@@ -845,14 +1163,15 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 static int osmo_stream_srv_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct osmo_stream_srv *conn = ofd->data;
+	int rc = 0;
 
 	LOGP(DLINP, LOGL_DEBUG, "connected read/write\n");
-	if (what & BSC_FD_READ)
-		osmo_stream_srv_read(conn);
-	if (what & BSC_FD_WRITE)
+	if (what & OSMO_FD_READ)
+		rc = osmo_stream_srv_read(conn);
+	if (rc != -EBADF && (what & OSMO_FD_WRITE))
 		osmo_stream_srv_write(conn);
 
-	return 0;
+	return rc;
 }
 
 /*! \brief Create a Stream Server inside the specified link
@@ -877,7 +1196,7 @@ osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link,
 	conn->ofd.fd = fd;
 	conn->ofd.data = conn;
 	conn->ofd.cb = osmo_stream_srv_cb;
-	conn->ofd.when = BSC_FD_READ;
+	conn->ofd.when = OSMO_FD_READ;
 	conn->cb = cb;
 	conn->closed_cb = closed_cb;
 	conn->data = data;
@@ -963,7 +1282,7 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 	}
 
 	msgb_enqueue(&conn->tx_queue, msg);
-	conn->ofd.when |= BSC_FD_WRITE;
+	conn->ofd.when |= OSMO_FD_WRITE;
 }
 
 /*! \brief Receive data via Osmocom stream server
